import traceback
from inspect import signature
from types import FunctionType
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse, urlencode
from typing import Union, Callable, List
from selenium.webdriver.chrome.options import Options
from .utils.logger import debug, info, warn, error
from .utils.url import get_query_params, get_domain
from .utils.chrome_driver import build_driver, get_websocket_debugger_url
from .utils.user_agent import get_random_user_agent
from .query import Query, QueryOptions
from .utils.constants import JOBS_SEARCH_URL
from .strategies import Strategy, AnonymousStrategy, AuthenticatedStrategy
from .config import Config
from .events import Events
from .chrome_cdp import CDP, CDPRequest, CDPResponse
from .exceptions import CallbackException, InvalidCookieException


class LinkedinScraper:
    """
    Args:
        chrome_options (selenium.webdriver.chrome.options.Options): Options to be passed to the Chrome driver.
            If None, default options will be used.
        headless (bool): Overrides headless mode only if chrome_options is None. If chrome_options is passed in
            the constructor, this flag is ignored.
        max_workers (int): Number of threads spawned to execute concurrent queries. Each thread will use a
            different Chrome driver instance.
        slow_mo (float): Slow down the scraper execution, mainly to avoid 429 (Too many requests) errors.
    """

    def __init__(
            self,
            chrome_executable_path = None,
            chrome_options: Options = None,
            headless: bool = True,
            max_workers: int = 2,
            slow_mo: float = 0.4):

        # Input validation
        if chrome_executable_path is not None and not isinstance(chrome_executable_path, str):
            raise ValueError('Input parameter chrome_executable_path must be of type str')

        if chrome_options is not None and not isinstance(chrome_options, Options):
            raise ValueError('Input parameter chrome_options must be instance of class '
                             'selenium.webdriver.chrome.options.Options')

        if not isinstance(max_workers, int) or max_workers < 1:
            raise ValueError('Input parameter max_workers must be a positive integer')

        if (not isinstance(slow_mo, int) and not isinstance(slow_mo, float)) or slow_mo < 0:
            raise ValueError('Input parameter slow_mo must be a positive number')

        self.chrome_executable_path = chrome_executable_path
        self.chrome_options = chrome_options
        self.headless = headless
        self.slow_mo = slow_mo
        self._pool = ThreadPoolExecutor(max_workers=max_workers)
        self._strategy: Strategy
        self._emitter = {
            Events.DATA: [],
            Events.ERROR: [],
            Events.INVALID_SESSION: [],
            Events.END: [],
        }

        if Config.LI_AT_COOKIE:
            info(f'Using strategy {AuthenticatedStrategy.__name__}')
            self._strategy = AuthenticatedStrategy(self)
        else:
            info(f'Using strategy {AnonymousStrategy.__name__}')
            self._strategy = AnonymousStrategy(self)

    @staticmethod
    def __build_search_url(query: Query, location: str = '') -> str:
        """
        Build jobs search url from query and location
        :param query: Query
        :param location: str
        :return: str
        """

        tag = f'[{query.query}][{location}]'
        parsed = urlparse(JOBS_SEARCH_URL)
        params = {}

        if len(query.query) > 0:
            params['keywords'] = query.query

        if len(location) > 0:
            params['location'] = location

        if query.options.filters is not None:
            if query.options.filters.company_jobs_url is not None:
                _params = get_query_params(query.options.filters.company_jobs_url)
                if 'f_C' in _params:
                    params['f_C'] = _params['f_C']
                    debug(tag, 'Applied company filter', query.options.filters.company_jobs_url)

            if query.options.filters.relevance is not None:
                params['sortBy'] = query.options.filters.relevance.value
                debug(tag, 'Applied relevance filter', query.options.filters.relevance)

            if query.options.filters.time is not None:
                params['f_TP' if not Config.LI_AT_COOKIE else 'f_TPR'] = query.options.filters.time.value
                debug(tag, 'Applied time filter', query.options.filters.time)

            if len(query.options.filters.type) > 0:
                filters = ','.join(e.value for e in query.options.filters.type)
                params['f_JT'] = filters
                debug(tag, 'Applied type filters', query.options.filters.type)

            if len(query.options.filters.experience) > 0:
                filters = ','.join(e.value for e in query.options.filters.experience)
                params['f_E'] = filters
                debug(tag, 'Applied experience filters', query.options.filters.experience)

        params['redirect'] = 'false'
        params['position'] = '1'
        params['pageNum'] = '0'

        parsed = parsed._replace(query=urlencode(params))
        return parsed.geturl()

    @staticmethod
    def __validate_run_input(queries: Union[Query, List[Query]], options: QueryOptions = None):
        """
        Validate run input parameters
        :param queries: Union[Query, List[Query]]
        :param options: QueryOptions
        :return: None
        """

        if queries is None:
            raise ValueError('Parameter queries is missing')

        if not isinstance(queries, list):
            queries = [queries]

        for query in queries:
            if not isinstance(query, Query):
                raise ValueError(f'A query object must be an instance of class Query, found {type(query)}')
            query.validate()

        if options is not None:
            if not isinstance(options, QueryOptions):
                raise ValueError(f'Parameter options must be an instance of class QueryOptions, found {type(options)}')
            options.validate()

    def __run(self, query: Query) -> None:
        """
        Run query in a new thread for each location
        :param query: Query
        :return: None
        """

        tag = f'[{query.query}]'
        driver = None
        devtools = None

        info('Starting new query', str(query))

        try:
            # Locations loop
            for location in query.options.locations:
                tag = f'[{query.query}][{location}]'
                search_url = LinkedinScraper.__build_search_url(query, location)

                driver = build_driver(
                    executable_path=self.chrome_executable_path,
                    options=self.chrome_options,
                    headless=self.headless
                )

                websocket_debugger_url = get_websocket_debugger_url(driver)
                devtools = CDP(websocket_debugger_url)

                def on_request(request: CDPRequest) -> None:
                    domain = get_domain(request.url)

                    # By default blocks all tracking and 3rd part domains requests
                    if 'li/track' in request.url or domain not in {'linkedin.com', 'licdn.com'}:
                        return request.abort()

                    # If optimize is enabled, blocks other resource types
                    if query.options.optimize:
                        types_to_block = {
                            'image',
                            'stylesheet',
                            'media',
                            'font',
                            'texttrack',
                            'object',
                            'beacon',
                            'csp_report',
                            'imageset',
                        }

                        if request.resource_type.lower() in types_to_block:
                            return request.abort()

                    request.resume()

                def on_response(response: CDPResponse) -> None:
                    if response.status == 429:
                        warn(tag, '[429] Too many requests', 'You should probably increase scraper "slow_mo" value '
                                                             'or reduce concurrency')
                    elif response.status >= 400:
                        warn(tag, 'Error in response', str(response))

                # Add request/response listeners
                devtools.on('request', on_request)
                devtools.on('response', on_response)

                # Start devtools
                devtools.start()

                # Set random user agent
                devtools.set_user_agent(get_random_user_agent())

                # Run strategy
                self._strategy.run(driver, search_url, query, location)

                try:
                    debug(tag, 'Stopping Chrome DevTools')
                    devtools.stop()
                except:
                    pass

                try:
                    debug(tag, 'Closing driver active window')
                    driver.close()
                except:
                    pass
        except CallbackException as e:
            error(tag, e)
            raise e
        except InvalidCookieException as e:
            error(tag, e)
            raise e
        except BaseException as e:
            error(tag, e)
            self.emit(Events.ERROR, str(e) + '\n' + traceback.format_exc())
        finally:
            try:
                debug(tag, 'Stopping Chrome DevTools')
                devtools.stop()
            except:
                pass

            try:
                debug(tag, 'Closing driver')
                driver.quit()
            except:
                pass

        # Emit END event
        self.emit(Events.END)

    def run(self, queries: Union[Query, List[Query]], options: QueryOptions = None) -> None:
        """
        Run a query or a list of queries
        :param queries: Union[Query, List[Query]]
        :param options: QueryOptions
        :return: None
        """

        # Validate input
        LinkedinScraper.__validate_run_input(queries, options)

        # Merge with global options
        global_options = options if options is not None \
            else QueryOptions(locations=['Worldwide'], limit=25, optimize=False)

        for query in queries:
            if not isinstance(query, Query):
                raise ValueError('A query must be instance of class Query')
            query.merge_options(global_options)

        futures = [self._pool.submit(self.__run, query) for query in queries]
        [f.result() for f in futures]  # Necessary also to get exceptions from futures

    def on(self, event: Events, cb: Callable, once=False) -> None:
        """
        Add callback for the given event
        :param event: str
        :param cb: Callable
        :param once: bool
        :return: None
        """

        if not isinstance(event, Events):
            raise ValueError(f'Event must be an instance of enum class Events')

        if not isinstance(cb, FunctionType):
            raise ValueError('Callback must be a function')

        if event == Events.DATA or event == Events.ERROR:
            allowed_params = 1
        else:
            allowed_params = 0

        if len(signature(cb).parameters) != allowed_params:
            raise ValueError(f'Callback for event {event} must have {allowed_params} arguments')

        self._emitter[event].append({'cb': cb, 'once': once})

    def once(self, event: Events, cb: Callable) -> None:
        """
        Add once callback for the given event
        :param event: str
        :param cb: Callable
        :return: None
        """

        self.on(event, cb, once=True)

    def emit(self, event: Events, *args) -> None:
        """
        Execute callbacks for the given event
        :param event: str
        :param args: args
        :return: None
        """

        if not isinstance(event, Events):
            raise ValueError(f'Event must be an instance of enum class Events')

        for listener in self._emitter[event]:
            try:
                listener['cb'](*args)
            except BaseException as e:
                raise CallbackException(str(e) + '\n' + traceback.format_exc())

        # Remove 'once' callbacks
        self._emitter[event] = [e for e in self._emitter[event] if not e['once']]

    def remove_listener(self, event: Events, cb: Callable) -> bool:
        """
        Remove listener for the given event
        :param event: str
        :param cb: Callable
        :return:
        """

        if not isinstance(event, Events):
            raise ValueError(f'Event must be an instance of enum class Events')

        n = len(self._emitter[event])
        self._emitter[event] = [e for e in self._emitter[event] if e['cb'] != cb]
        return len(self._emitter[event]) < n

    def remove_all_listeners(self, event: Events) -> None:
        """
        Remove all listeners for the given event
        :param event: str
        :return: None
        """

        if not isinstance(event, Events):
            raise ValueError(f'Event must be an instance of enum class Events')

        self._emitter[event] = []



import os
import logging


class Config:
    LI_AT_COOKIE = os.environ['LI_AT_COOKIE'] if 'LI_AT_COOKIE' in os.environ else None
    LOGGER_NAMESPACE = 'li:scraper'

    _level = logging.INFO

    if 'LOG_LEVEL' in os.environ:
        _level_env = os.environ['LOG_LEVEL'].upper().strip()

        if _level_env == 'DEBUG':
            _level = logging.DEBUG
        elif _level_env == 'INFO':
            _level = logging.INFO
        elif _level_env == 'WARN' or _level_env == 'WARNING':
            _level = logging.WARN
        elif _level_env == 'ERROR':
            _level = logging.ERROR
        elif _level_env == 'FATAL':
            _level = logging.FATAL

    LOGGER_LEVEL = _level



from typing import List, Union
from ..filters import TimeFilters, ExperienceLevelFilters, TypeFilters, RelevanceFilters
from ..utils.url import get_query_params


class __Base:
    def __str__(self):
        params = [f'{k}={str(v)}' for k, v in self.__dict__.items() if v is not None and not self.__is_empty_list(v)]
        return f'{self.__class__.__name__}({" ".join(params)})'

    @staticmethod
    def __is_empty_list(v):
        return isinstance(v, List) and len(v) == 0


class QueryFilters(__Base):
    def __init__(self,
                 company_jobs_url: str = None,
                 relevance: RelevanceFilters = None,
                 time: TimeFilters = None,
                 type: Union[TypeFilters, List[TypeFilters]] = None,
                 experience: Union[ExperienceLevelFilters, List[ExperienceLevelFilters]] = None):

        super().__init__()

        if type is not None:
            if not isinstance(type, List):
                type = [type]
        else:
            type = []

        if experience is not None:
            if not isinstance(experience, List):
                experience = [experience]
        else:
            experience = []

        self.company_jobs_url = company_jobs_url
        self.relevance = relevance
        self.time = time
        self.type = type
        self.experience = experience

    def validate(self):
        if self.company_jobs_url is not None:
            if not isinstance(self.company_jobs_url, str):
                raise ValueError('Parameter company_jobs_url must be a string')

            try:
                query_params = get_query_params(self.company_jobs_url)
                if 'f_C' not in query_params:
                    raise ValueError('Parameter company_jobs_url is invalid. '
                                     'Please check the documentation on how find a company jobs link from LinkedIn')
            except:
                raise ValueError('Parameter company_jobs_url must be a valid url')

        if self.relevance is not None and not isinstance(self.relevance, RelevanceFilters):
            raise ValueError('Parameter relevance must be of type RelevanceFilters')

        if self.time is not None and not isinstance(self.time, TimeFilters):
            raise ValueError('Parameter time must be of type TimeFilters')

        if any((not isinstance(e, TypeFilters) for e in self.type)):
            raise ValueError('Parameter type must be of type Union[TypeFilters, List[TypeFilters]]')

        if any((not isinstance(e, ExperienceLevelFilters) for e in self.experience)):
            raise ValueError('Parameter experience must be of type '
                             'Union[ExperienceLevelFilters, List[ExperienceLevelFilters]]')


class QueryOptions(__Base):
    def __init__(self,
                 limit: int = None,
                 locations: List[str] = None,
                 filters: QueryFilters = None,
                 optimize: bool = None):

        super().__init__()

        if isinstance(locations, str):
            locations = [locations]

        self.limit = limit
        self.locations = locations
        self.filters = filters
        self.optimize = optimize

    def validate(self):
        if self.limit is not None:
            if not isinstance(self.limit, int) or self.limit < 0:
                raise ValueError('Parameter limit must be a positive integer')

        if self.locations is not None:
            if not isinstance(self.locations, List) or any([not isinstance(e, str) for e in self.locations]):
                raise ValueError('Parameter locations must be a list of strings')

        if self.optimize is not None and not isinstance(self.optimize, bool):
            raise ValueError('Parameter optimize must be a boolean')

        if self.filters is not None:
            self.filters.validate()


class Query(__Base):
    def __init__(self, query: str = '', options: QueryOptions = QueryOptions()):
        super().__init__()

        self.query = query
        self.options = options

    def merge_options(self, options: QueryOptions):
        if self.options.limit is None:
            self.options.limit = options.limit if options.limit is not None else 25

        if self.options.optimize is None:
            self.options.optimize = options.optimize if options.optimize is not None else False

        if self.options.locations is None and options.locations is not None:
            self.options.locations = options.locations

        if self.options.filters is None and options.filters is not None:
            self.options.filters = options.filters

    def validate(self):
        if not isinstance(self.query, str):
            raise ValueError(f'Parameter query must be a string')

        self.options.validate()



class CDPResponse:
    def __init__(self, parent: 'ChromeDevTools', message):
        self._parent = parent
        self._tag = '[CDPResponse]'
        params = message['params']
        response = params['response']

        self.event: str = message.pop('method', None)
        self.request_id: str = params.pop('requestId', None)
        self.loader_id: str = params.pop('loaderId', None)
        self.timestamp: float = params.pop('timestamp', None)
        self.type: str = params.pop('type', None)
        self.frame_id: str = params.pop('frameId', None)
        self.url: str = response.pop('url', None)
        self.status: int = response.pop('status', None)
        self.status_text: str = response.pop('statusText', None)
        self.headers: dict = response.pop('headers', None)
        self.mime_type = response.pop('mimeType', None)
        self.connection_reused: bool = response.pop('connectionReused', None)
        self.connection_id: int = response.pop('connectionId', None)
        self.remote_ip_address: str = response.pop('remoteIPAddress', None)
        self.remote_port: int = response.pop('remotePort', None)
        self.from_disk_cache: bool = response.pop('fromDiskCache', None)
        self.from_service_worker: bool = response.pop('fromServiceWorker', None)
        self.from_prefetch_cache: bool = response.pop('fromPrefetchCache', None)
        self.encoded_data_length: int = response.pop('encodedDataLength', None)
        self.timing: dict = response.pop('timing', None)
        self.response_time: float = response.pop('responseTime', None)
        self.protocol: str = response.pop('protocol', None)
        self.security_state: str = response.pop('securityState', None)
        self.security_details: dict = response.pop('securityDetails', None)

    def __str__(self):
        return f'request_id={self.request_id} status={self.status} type={self.type} mime_type={self.mime_type} url={self.url}'



from .strategy import Strategy
from .authenticated_strategy import AuthenticatedStrategy
from .anonymous_strategy import AnonymousStrategy

__all__ = [
    'Strategy',
    'AnonymousStrategy',
    'AuthenticatedStrategy',
]



import traceback
from typing import NamedTuple
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from urllib.parse import urlparse
from time import sleep
from .strategy import Strategy
from ..query import Query
from ..utils.logger import debug, info, warn, error
from ..events import Events, EventData


class Selectors(NamedTuple):
    container = '.results__container.results__container--two-pane'
    jobs = '.jobs-search__results-list li'
    links = '.jobs-search__results-list li a.result-card__full-card-link'
    applyLink = 'a[data-is-offsite-apply=true]'
    dates = 'time'
    companies = '.result-card__subtitle.job-result-card__subtitle'
    places = '.job-result-card__location'
    detailsTop = '.topcard__content-left'
    description = '.description__text'
    criteria = 'li.job-criteria__item'
    seeMoreJobs = 'button.infinite-scroller__show-more-button'


class AnonymousStrategy(Strategy):
    def __init__(self, scraper: 'LinkedinScraper'):
        super().__init__(scraper)

    @staticmethod
    def __require_authentication(driver: webdriver) -> bool:
        """
        Verify if driver has been redirected to auth wall and needs authentication
        :param driver: webdriver
        :return: bool
        """

        parsed = urlparse(driver.current_url)
        return 'authwall' in parsed.path.lower()

    @staticmethod
    def __load_job_details(driver: webdriver, timeout=2) -> object:
        """
        Wait for job details to load
        :param driver: webdriver
        :param timeout: int
        :return: object
        """
        elapsed = 0
        sleep_time = 0.05

        while elapsed < timeout:
            loaded = driver.execute_script(
                '''
                    const description = document.querySelector(arguments[0]);
                    return description && description.innerText.length > 0;    
                ''',
                Selectors.description)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on loading job details'}

    @staticmethod
    def __load_more_jobs(driver: webdriver, job_links_tot: int, timeout=2) -> object:
        """

        :param driver:
        :param job_links_tot:
        :param timeout:
        :return:
        """

        elapsed = 0
        sleep_time = 0.05
        clicked = False

        while elapsed < timeout:
            if not clicked:
                clicked = driver.execute_script(
                    '''
                        const button = document.querySelector(arguments[0]);
    
                        if (button) {
                            button.click();
                            return true;
                        }
                        else {
                            return false;
                        }    
                    ''',
                    Selectors.seeMoreJobs)

            loaded = driver.execute_script(
                '''
                    window.scrollTo(0, document.body.scrollHeight);
                    return document.querySelectorAll(arguments[0]).length > arguments[1];
                ''',
                Selectors.links,
                job_links_tot)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on loading more jobs'}

    def run(self, driver: webdriver, search_url: str, query: Query, location: str) -> None:
        """
        Run scraper
        :param driver: webdriver
        :param search_url: str
        :param query: Query
        :param location: str
        :return: None
        """

        tag = f'[{query.query}][{location}]'
        processed = 0

        info(tag, f'Opening {search_url}')
        driver.get(search_url)

        # Verify if redirected to auth wall
        if AnonymousStrategy.__require_authentication(driver):
            error('Scraper failed to run in anonymous mode, authentication may be necessary for this environment. '
                  'Please check the documentation on how to use an authenticated session.')
            return

        # Wait container
        try:
            WebDriverWait(driver, 5).until(ec.presence_of_element_located((By.CSS_SELECTOR, Selectors.container)))
        except BaseException as e:
            info(tag, 'No jobs found, skip')
            return

        job_index = 0

        # Pagination loop
        while processed < query.options.limit:
            job_links_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;',
                                                  Selectors.links)

            if job_links_tot == 0:
                info(tag, 'No jobs found, skip')
                break

            info(tag, f'Found {job_links_tot} jobs')

            # Jobs loop
            while job_index < job_links_tot and processed < query.options.limit:
                sleep(self.scraper.slow_mo)
                tag = f'[{query.query}][{location}][{processed + 1}]'

                # Extract job main fields
                debug(tag, 'Evaluating selectors', [
                    Selectors.links,
                    Selectors.companies,
                    Selectors.places,
                    Selectors.dates])

                try:
                    job_id, job_title, job_company, job_place, job_date = driver.execute_script(
                        '''
                            return [
                                document.querySelectorAll(arguments[1])[arguments[0]].getAttribute('data-id'),
                                document.querySelectorAll(arguments[2])[arguments[0]].innerText,
                                document.querySelectorAll(arguments[3])[arguments[0]].innerText,
                                document.querySelectorAll(arguments[4])[arguments[0]].innerText,
                                document.querySelectorAll(arguments[5])[arguments[0]].getAttribute('datetime')
                            ];
                        ''',
                        job_index,
                        Selectors.jobs,
                        Selectors.links,
                        Selectors.companies,
                        Selectors.places,
                        Selectors.dates)

                    # Load job details and extract job link
                    debug(tag, 'Evaluating selectors', [
                        Selectors.links])

                    job_link = driver.execute_script(
                        '''
                            const linkElem = document.querySelectorAll(arguments[1])[arguments[0]];
                            linkElem.scrollIntoView();
                            linkElem.click();
                            return linkElem.getAttribute("href");
                        ''',
                        job_index,
                        Selectors.links)

                    # Wait for job details to load
                    load_result = AnonymousStrategy.__load_job_details(driver)

                    if not load_result['success']:
                        error(tag, load_result['error'])
                        job_index += 1
                        continue

                    # Exctract
                    debug(tag, 'Evaluating selectors', [Selectors.description])

                    job_description, job_description_html = driver.execute_script(
                        '''
                            const el = document.querySelector(arguments[0]);
                        
                            return [
                                el.innerText,
                                el.outerHTML    
                            ];
                        ''',
                        Selectors.description)

                    # Extract apply link
                    debug(tag, 'Evaluating selectors', [Selectors.applyLink])

                    job_apply_link = driver.execute_script(
                        '''
                            const applyBtn = document.querySelector(arguments[0]);
                            return applyBtn ? applyBtn.getAttribute("href") : '';
                        ''',
                        Selectors.applyLink)

                    # Extract criteria
                    debug(tag, 'Evaluating selectors', [Selectors.criteria])

                    job_seniority_level, job_function, job_employment_type, job_industries = driver.execute_script(
                        '''
                            const items = document.querySelectorAll(arguments[0]);
    
                            const criteria = [
                                'Seniority level',
                                'Job function',
                                'Employment type',
                                'Industries'
                            ];
    
                            const nodeList = criteria.map(criteria => {
                                const el = Array.from(items)
                                    .find(li =>
                                        (li.querySelector('h3')).innerText === criteria);
    
                                return el ? el.querySelectorAll('span') : [];
                            });
    
                            return Array.from(nodeList)
                                .map(spanList => Array.from(spanList)
                                    .map(e => e.innerText).join(', '));
                        ''',
                        Selectors.criteria)

                except BaseException as e:
                    error(tag, e, traceback.format_exc())
                    self.scraper.emit(Events.ERROR, str(e) + '\n' + traceback.format_exc())
                    job_index += 1
                    continue

                data = EventData(
                    query=query.query,
                    location=location,
                    job_id=job_id,
                    job_index=job_index,
                    title=job_title,
                    company=job_company,
                    place=job_place,
                    date=job_date,
                    link=job_link,
                    apply_link=job_apply_link,
                    description=job_description,
                    description_html=job_description_html,
                    seniority_level=job_seniority_level,
                    job_function=job_function,
                    employment_type=job_employment_type,
                    industries=job_industries)

                info(tag, 'Processed')

                job_index += 1
                processed += 1

                self.scraper.emit(Events.DATA, data)

                # Try fetching more jobs
                if processed < query.options.limit and job_index == job_links_tot:
                    job_links_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;',
                                                          Selectors.links)

            # Check if we reached the limit of jobs to process
            if processed == query.options.limit:
                break

            # Check if we need to paginate
            info(tag, 'Checking for new jobs to load...')
            load_result = AnonymousStrategy.__load_more_jobs(driver, job_links_tot)

            if not load_result['success']:
                info(tag, "Couldn't find more jobs for the running query")
                break



import threading
import json
import websocket
from websocket import WebSocketTimeoutException, WebSocketConnectionClosedException
from types import FunctionType
from typing import Callable, Union
from .request import CDPRequest
from .response import CDPResponse
from .events import Events
from ..utils.logger import debug, info, warn, error


class CDP:
    def __init__(self, ws_url, timeout=1):
        self.ws_url = ws_url
        self.timeout = timeout
        self._tag = f'[T_{threading.get_ident()} ChromeDevTools]'
        self._stop = threading.Event()
        self._ws = None
        self._ws_loop_th = None
        self._is_running = False
        self._id = 0

        def __default_request_handler(request: CDPRequest) -> None:
            request.resume()

        self._event_handlers = {
            'request': __default_request_handler,
            'response': None
        }

    def __ws_loop(self):
        """
        Loop to receive messages from websocket server
        :return:
        """

        while not self._stop.is_set():
            try:
                msg = self._ws.recv()
                parsed = json.loads(msg)

                # Intercept request/response
                if 'method' in parsed:
                    event = parsed['method']

                    # Request handler
                    if event == Events.REQUEST.value:
                        if self._event_handlers['request'] is not None:
                            request = CDPRequest(self, parsed)
                            self._event_handlers['request'](request)

                    # Response handler
                    if event == Events.RESPONSE.value:
                        if self._event_handlers['response'] is not None:
                            response = CDPResponse(self, parsed)
                            self._event_handlers['response'](response)
            except (WebSocketTimeoutException, WebSocketConnectionClosedException) as e:
                continue

    def call_method(self, method: str, **params) -> None:
        """
        Call dev tools method with the given parameters
        :param method: str
        :param params:
        :return: None
        """

        if not self._ws or not self._ws.connected:
            raise RuntimeError(self._tag + '\tWebsocket not connected')

        self._id += 1

        msg = {'id': self._id, 'method': method, 'params': params}
        debug(self._tag, 'Calling method', msg)
        self._ws.send(json.dumps(msg))

    def start(self) -> None:
        """
        Start ChromeDevTools client
        :return: None
        """

        if self._is_running:
            raise RuntimeError(self._tag, 'It is already running')

        debug(self._tag, 'Connecting to websocket', self.ws_url)
        self._ws = websocket.create_connection(self.ws_url, enable_multithread=True, skip_utf8_validation=True)
        self._ws.settimeout(self.timeout)

        # Enable Fetch domain
        self.call_method('Fetch.enable')

        # Enable Network domain
        self.call_method('Network.enable')

        self._stop.clear()
        self._ws_loop_th = threading.Thread(target=self.__ws_loop, daemon=True)
        debug(self._tag, 'Starting websocket loop thread', self._ws_loop_th.ident)
        self._ws_loop_th.start()
        self._is_running = True

    def stop(self) -> None:
        """
        Stop ChromeDevTools client
        :return: None
        """

        self._stop.set()

        if self._ws_loop_th:
            self._ws_loop_th.join()

        self._is_running = False

        debug(self._tag, 'Closing websocket')
        self._ws.close()

    def on(self, event: str, cb: Union[Callable, None]) -> None:
        """
        Override event handler
        :param event: str
        :param cb: Callable | None
        :return: None
        """

        if cb is not None and not isinstance(cb, FunctionType):
            raise ValueError(self._tag + '\tCallback must be a function')

        if event not in self._event_handlers.keys():
            raise ValueError(self._tag + f'\tEvent must be one of ({", ".join(self._event_handlers.keys())})')

        self._event_handlers[event] = cb

    def set_user_agent(self, ua: str) -> None:
        """
        Set user agent
        :param ua: str
        :return: None
        """

        debug(self._tag, 'Setting user agent', ua)
        self.call_method('Network.setUserAgentOverride', userAgent=ua)



from enum import Enum
from ..config import Config


class RelevanceFilters(Enum):
    RELEVANT = 'R'
    RECENT = 'DD'


class TimeFilters(Enum):
    ANY = ''
    DAY = '1' if not Config.LI_AT_COOKIE else 'r86400'
    WEEK = '1,2' if not Config.LI_AT_COOKIE else 'r604800'
    MONTH = '1,2,3,4' if not Config.LI_AT_COOKIE else 'r2592000'


class TypeFilters(Enum):
    FULL_TIME = 'F'
    PART_TIME = 'P'
    TEMPORARY = 'T'
    CONTRACT = 'C'
    INTERNSHIP = 'I'
    VOLUNTEER = 'V'
    OTHER = 'O'


class ExperienceLevelFilters(Enum):
    INTERNSHIP = '1'
    ENTRY_LEVEL = '2'
    ASSOCIATE = '3'
    MID_SENIOR = '4'
    DIRECTOR = '5'
    EXECUTIVE = '6'



from selenium import webdriver
from ..query import Query


class Strategy:
    def __init__(self, scraper: 'LinkedinScraper'):
        self.scraper = scraper

    def run(self, driver: webdriver, search_url: str, query: Query, location: str) -> None:
        raise NotImplementedError('Must implement method in subclass')



class CallbackException(Exception):
    """Raised when an exception occurs in a scraper callback"""

    def __init__(self, *args):
        super().__init__(*args)


class InvalidCookieException(Exception):
    """Raised when the session cookie is invalid"""

    def __init__(self, *args):
        super().__init__(*args)


import os
import logging
from ..config import Config

logger = logging.getLogger(Config.LOGGER_NAMESPACE)
logger.setLevel(Config.LOGGER_LEVEL)


def __format(*args):
    return '\t'.join([str(arg) for arg in args])


def debug(*args):
    logger.debug(__format(args))


def info(*args):
    logger.info(__format(args))


def warn(*args):
    logger.warning(__format(args))


def error(*args):
    logger.error(__format(args), exc_info=True)



from ..utils.logger import debug


class CDPRequest:
    def __init__(self, parent: 'ChromeDevTools', message):
        self._parent = parent
        self._tag = '[CDPRequest]'
        params = message['params']
        request = params['request']

        self.event: str = message.pop('method', None)
        self.request_id: str = params.pop('requestId', None)
        self.url: str = request.pop('url', None)
        self.method: str = request.pop('method', None)
        self.headers: dict = request.pop('headers', None)
        self.post_data: dict = request.pop('postData', None)
        self.initial_priority: str = request.pop('initialPriority', None)
        self.referrer_policy: str = request.pop('referrerPolicy', None)
        self.resource_type: str = params.pop('resourceType', None)
        self.frame_id: str = params.pop('frameId', None)

    def __str__(self):
        return f'request_id={self.request_id} method={self.method} resource_type={self.resource_type} url={self.url}'

    def resume(self):
        debug(self._tag, '[RESUME]', str(self))
        self._parent.call_method('Fetch.continueRequest', requestId=self.request_id)

    def abort(self, reason='Aborted'):
        debug(self._tag, '[ABORT]', str(self))
        self._parent.call_method('Fetch.failRequest', requestId=self.request_id, errorReason=reason)



from random import choice

_user_agents = [
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2859.0 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.49 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:55.0) Gecko/20100101 Firefox/55.0',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2876.0 Safari/537.36',
    'Mozilla/5.0 (X11; Linux i686 (x86_64)) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3187.0 Safari/537.366',
    'Mozilla/5.0 (X11; Fedora; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/62.0.3178.0 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:55.0) Gecko/20100101 Firefox/55.0',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:55.0) Gecko/20100101 Firefox/55.0',
]


def get_random_user_agent() -> str:
    return choice(_user_agents)



from enum import Enum
from typing import NamedTuple


class Events(Enum):
    DATA = 'scraper:data'
    END = 'scraper:end'
    ERROR = 'scraper:error'
    INVALID_SESSION = 'scraper:invalid-session'


class EventData(NamedTuple):
    query: str = ''
    location: str = ''
    job_id: str = ''
    job_index: int = -1  # Only for debug
    link: str = ''
    apply_link: str = ''
    title: str = ''
    company: str = ''
    place: str = ''
    description: str = ''
    description_html: str = ''
    date: str = ''
    seniority_level: str = ''
    job_function: str = ''
    employment_type: str = ''
    industries: str = ''



from setuptools import setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
    name='linkedin-jobs-scraper',
    version='1.5.2',
    author='Ludovico Fabbri',
    author_email='ludovico.fabbri@gmail.com',
    description='Scrape public available jobs on Linkedin using headless browser',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/spinlud/py-linkedin-jobs-scraper.git',
    packages=[
        'linkedin_jobs_scraper',
        'linkedin_jobs_scraper.chrome_cdp',
        'linkedin_jobs_scraper.events',
        'linkedin_jobs_scraper.exceptions',
        'linkedin_jobs_scraper.filters',
        'linkedin_jobs_scraper.query',
        'linkedin_jobs_scraper.strategies',
        'linkedin_jobs_scraper.utils',
    ],
    install_requires=[
        'selenium',
        'websocket-client'
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import os
import traceback
from typing import NamedTuple
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as ec
from time import sleep
from .strategy import Strategy
from ..config import Config
from ..query import Query
from ..utils.logger import debug, info, warn, error
from ..utils.constants import HOME_URL
from ..events import Events, EventData
from ..exceptions import InvalidCookieException


class Selectors(NamedTuple):
    container = '.jobs-search-two-pane__container'
    chatPanel = '.msg-overlay-list-bubble'
    jobs = '.job-card-container'
    links = 'a.job-card-container__link.job-card-list__title'
    companies = '.job-card-container .artdeco-entity-lockup__subtitle'  # OK
    places = '.job-card-container .artdeco-entity-lockup__caption'  # OK
    dates = '.job-card-container time'  # OK
    description = '.jobs-description'
    detailsTop = '.jobs-details-top-card'
    details = '.jobs-details__main-content'
    criteria = '.jobs-box__group h3'
    pagination = '.jobs-search-two-pane__pagination'
    paginationBtn = lambda index: f'li[data-test-pagination-page-btn="{index}"] button'


class AuthenticatedStrategy(Strategy):
    def __init__(self, scraper: 'LinkedinScraper'):
        super().__init__(scraper)

    @staticmethod
    def __is_authenticated_session(driver: webdriver):
        """
        Return True if authenticated session cookie is set, False otherwise
        :param driver: webdriver
        :return:
        """
        return driver.get_cookie('li_at') is not None

    @staticmethod
    def __load_job_details(driver: webdriver, timeout=2) -> object:
        """
        Wait for job details to load
        :param driver: webdriver
        :param timeout: int
        :return: object
        """
        elapsed = 0
        sleep_time = 0.05

        while elapsed < timeout:
            loaded = driver.execute_script(
                '''
                    const description = document.querySelector(arguments[0]);
                    return description && description.innerText.length > 0;    
                ''',
                Selectors.description)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on loading job details'}

    @staticmethod
    def __paginate(driver: webdriver, pagination_index: int, timeout=2) -> object:
        """
        Attempt to paginate
        :param driver: webdriver
        :param pagination_index: int
        :return: object
        """

        elapsed = 0
        sleep_time = 0.05
        clicked = False
        pagination_btn_selector = Selectors.paginationBtn(pagination_index)

        # Wait pagination html to load
        try:
            WebDriverWait(driver, 5).until(ec.presence_of_element_located((By.CSS_SELECTOR, Selectors.pagination)))
        except:
            return {'success': False, 'error': 'Timeout on pagination'}

        # Try click next pagination button (if exists)
        clicked = driver.execute_script(
            '''
                const btn = document.querySelector(arguments[0]);

                if (btn) {
                    btn.click();
                    return true;
                }
                else {
                    return false;
                }
            ''',
            pagination_btn_selector)

        # Failed to click next pagination button (pagination exhausted)
        if not clicked:
            return {'success': False, 'error': 'Pagination exhausted'}

        # Wait for new jobs to load
        while elapsed < timeout:
            loaded = driver.execute_script(
                '''
                    return document.querySelectorAll(arguments[0]).length > 0;                
                ''',
                Selectors.links)

            if loaded:
                return {'success': True}

            sleep(sleep_time)
            elapsed += sleep_time

        return {'success': False, 'error': 'Timeout on pagination'}

    def run(self, driver: webdriver, search_url: str, query: Query, location: str) -> None:
        """
        Run strategy
        :param driver: webdriver
        :param search_url: str
        :param query: Query
        :param location: str
        :return: None
        """

        tag = f'[{query.query}][{location}]'
        processed = 0
        pagination_index = 1

        # Open main page first to verify/set the session
        debug(tag, f'Opening {HOME_URL}')
        driver.get(HOME_URL)
        sleep(self.scraper.slow_mo)

        if not AuthenticatedStrategy.__is_authenticated_session(driver):
            info(tag, 'Setting authentication cookie')

            try:
                driver.add_cookie({
                    'name': 'li_at',
                    'value': Config.LI_AT_COOKIE,
                    'domain': '.www.linkedin.com'
                })

                # driver.get(search_url)  # Need to open url again after setting cookie
                # sleep(self.scraper.slow_mo)
            except BaseException as e:
                error(tag, e)
                error(tag, traceback.format_exc())
                return

        # Open search url
        info(tag, f'Opening {search_url}')
        driver.get(search_url)
        sleep(self.scraper.slow_mo)

        # Verify session
        if not AuthenticatedStrategy.__is_authenticated_session(driver):
            message = 'The provided session cookie is invalid. ' \
                      'Check the documentation on how to obtain a valid session cookie.'
            raise InvalidCookieException(message)

        # Wait container
        try:
            WebDriverWait(driver, 5).until(ec.presence_of_element_located((By.CSS_SELECTOR, Selectors.container)))
        except BaseException as e:
            warn(tag, 'No jobs found, skip')
            return

        # Try closing chat panel
        try:
            driver.execute_script(
                '''
                    const div = document.querySelector(arguments[0]);
                    if (div) {
                        div.style.display = "none";
                    }                
                ''',
                Selectors.chatPanel)
        except:
            pass

        # Pagination loop
        while processed < query.options.limit:
            # Verify session in loop
            if not AuthenticatedStrategy.__is_authenticated_session(driver):
                warn(tag, 'Session is no longer valid, this may cause the scraper to fail')
                self.scraper.emit(Events.INVALID_SESSION)
            else:
                info(tag, 'Session is valid')

            job_index = 0

            job_links_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;',
                                                  Selectors.links)

            if job_links_tot == 0:
                info(tag, 'No jobs found, skip')
                break

            info(tag, f'Found {job_links_tot} jobs')

            # Jobs loop
            while job_index < job_links_tot and processed < query.options.limit:
                sleep(self.scraper.slow_mo)
                tag = f'[{query.query}][{location}][{processed + 1}]'

                # Extract job main fields
                debug(tag, 'Evaluating selectors', [
                    Selectors.links,
                    Selectors.companies,
                    Selectors.places,
                    Selectors.dates])

                try:
                    job_id, job_title, job_company, job_place, job_date = driver.execute_script(
                        '''
                            const jobId = document.querySelectorAll(arguments[1])[arguments[0]] ?
                                document.querySelectorAll(arguments[1])[arguments[0]].getAttribute("data-job-id") : "";
                
                            const title = document.querySelectorAll(arguments[2])[arguments[0]] ?
                                document.querySelectorAll(arguments[2])[arguments[0]].innerText : "";

                            const company = document.querySelectorAll(arguments[3])[arguments[0]] ?
                                document.querySelectorAll(arguments[3])[arguments[0]].innerText : "";

                            const place = document.querySelectorAll(arguments[4])[arguments[0]] ?
                                document.querySelectorAll(arguments[4])[arguments[0]].innerText : "";

                            const date = document.querySelectorAll(arguments[5])[arguments[0]] ?
                                document.querySelectorAll(arguments[5])[arguments[0]].getAttribute('datetime') : "";

                            return [
                                jobId,
                                title,
                                company,
                                place,
                                date,
                            ];                                                    
                        ''',
                        job_index,
                        Selectors.jobs, Selectors.links,
                        Selectors.companies,
                        Selectors.places,
                        Selectors.dates)

                    # Load job details and extract job link
                    debug(tag, 'Evaluating selectors', [
                        Selectors.links])

                    job_link = driver.execute_script(
                        '''
                            const linkElem = document.querySelectorAll(arguments[1])[arguments[0]];
                            linkElem.scrollIntoView();
                            linkElem.click();
                            return linkElem.getAttribute("href");
                        ''',
                        job_index, Selectors.links)

                    sleep(self.scraper.slow_mo)

                    # Wait for job details to load
                    load_result = AuthenticatedStrategy.__load_job_details(driver)

                    if not load_result['success']:
                        error(tag, load_result['error'])
                        job_index += 1
                        continue

                    # Extract
                    debug(tag, 'Evaluating selectors', [Selectors.description])

                    job_description, job_description_html = driver.execute_script(
                        '''
                            const el = document.querySelector(arguments[0]);

                            return [
                                el.innerText,
                                el.outerHTML    
                            ];
                        ''',
                        Selectors.description)

                    # TODO how to extract apply link?

                    # Extract criteria
                    debug(tag, 'Evaluating selectors', [Selectors.criteria])

                    job_seniority_level, job_function, job_employment_type, job_industries = driver.execute_script(
                        r'''
                            const nodes = document.querySelectorAll(arguments[0]);

                            const criteria = [
                                "Seniority Level",
                                "Employment Type",
                                "Industry",
                                "Job Functions",
                            ];

                            return Array.from(criteria.map(c => {
                                const el = Array.from(nodes).find(node => node.innerText.trim() === c);

                                if (el && el.nextElementSibling) {
                                    const sibling = el.nextElementSibling;
                                    return sibling.innerText
                                        .replace(/[\s]{2,}/g, ", ")
                                        .replace(/[\n\r]+/g, " ")
                                        .trim();
                                }
                                else {
                                    return "";
                                }
                            }));
                        ''',
                        Selectors.criteria)

                except BaseException as e:
                    error(tag, e, traceback.format_exc())
                    self.scraper.emit(Events.ERROR, str(e) + '\n' + traceback.format_exc())
                    job_index += 1
                    continue

                data = EventData(
                    query=query.query,
                    location=location,
                    job_id=job_id,
                    job_index=job_index,
                    title=job_title,
                    company=job_company,
                    place=job_place,
                    date=job_date,
                    link=job_link,
                    apply_link='',
                    description=job_description,
                    description_html=job_description_html,
                    seniority_level=job_seniority_level,
                    job_function=job_function,
                    employment_type=job_employment_type,
                    industries=job_industries)

                info(tag, 'Processed')

                job_index += 1
                processed += 1

                self.scraper.emit(Events.DATA, data)

                # Try fetching more jobs
                if processed < query.options.limit and job_index == job_links_tot:
                    job_links_tot = driver.execute_script('return document.querySelectorAll(arguments[0]).length;',
                                                          Selectors.links)

            # Check if we reached the limit of jobs to process
            if processed == query.options.limit:
                break

            # Try to paginate
            pagination_index += 1
            info(tag, f'Pagination requested ({pagination_index})')
            paginate_result = AuthenticatedStrategy.__paginate(driver, pagination_index)

            if not paginate_result['success']:
                info(tag, "Couldn't find more jobs for the running query")
                return



import urllib3
import json
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from linkedin_jobs_scraper.utils.logger import debug


def get_default_driver_options(width=1472, height=828, headless=True) -> Options:
    """
    Generate default Chrome driver options
    :param width: int
    :param height: int
    :param headless: bool
    :return: Options
    """

    chrome_options = Options()
    chrome_options.headless = headless
    chrome_options.page_load_strategy = 'normal'

    chrome_options.add_argument('--enable-automation'),
    chrome_options.add_argument('--start-maximized'),
    chrome_options.add_argument(f'--window-size={width},{height}'),
    chrome_options.add_argument('--lang=en-GB'),
    chrome_options.add_argument('--no-sandbox'),
    chrome_options.add_argument('--disable-setuid-sandbox'),
    chrome_options.add_argument('--disable-gpu'),
    chrome_options.add_argument('--disable-dev-shm-usage'),
    chrome_options.add_argument('--no-sandbox'),
    chrome_options.add_argument('--disable-setuid-sandbox'),
    chrome_options.add_argument('--disable-dev-shm-usage'),
    chrome_options.add_argument("-proxy-server='direct://"),
    chrome_options.add_argument('--proxy-bypass-list=*'),
    chrome_options.add_argument('--disable-accelerated-2d-canvas'),
    chrome_options.add_argument('--disable-gpu'),
    chrome_options.add_argument('--allow-running-insecure-content'),
    chrome_options.add_argument('--disable-web-security'),
    chrome_options.add_argument('--disable-client-side-phishing-detection'),
    chrome_options.add_argument('--disable-notifications'),
    chrome_options.add_argument('--mute-audio'),

    # Disable downloads
    chrome_options.add_experimental_option(
        'prefs', {
            'safebrowsing.enabled': 'false',
            'download.prompt_for_download': False,
            'download.default_directory': '/dev/null',
            'download_restrictions': 3,
            'profile.default_content_setting_values.notifications': 2,
        }
    )

    return chrome_options


def build_driver(executable_path: str = None, options: Options = None, headless=True, timeout=20) -> webdriver:
    """
    Build Chrome driver instance
    :param executable_path: str
    :param options: Options
    :param headless: bool
    :param timeout: int
    :return: webdriver
    """

    kwargs = {}

    if executable_path is not None:
        kwargs['executable_path'] = executable_path

    kwargs['options'] = options if options is not None else get_default_driver_options(headless=headless)

    driver = webdriver.Chrome(**kwargs)
    driver.set_page_load_timeout(timeout)

    return driver


def get_debugger_url(driver: webdriver) -> str:
    """
    Get Chrome debugger url
    :param driver: webdriver
    :return: str
    """

    chrome_debugger_url = f"http://{driver.capabilities['goog:chromeOptions']['debuggerAddress']}"
    debug('Chrome Debugger Url', chrome_debugger_url)
    return chrome_debugger_url


def get_websocket_debugger_url(driver: webdriver) -> str:
    """
    Get Chrome websocket debugger url
    :param driver: webdriver
    :return: str
    """

    chrome_debugger_url = get_debugger_url(driver)
    http = urllib3.PoolManager()
    response = json.loads(http.request('GET', chrome_debugger_url + '/json').data.decode())
    return response[0]['webSocketDebuggerUrl']



from urllib.parse import urlparse, urlencode, parse_qsl


def get_query_params(url: str) -> dict:
    """
    Extract url query parameters as a dictionary
    :param url:
    :return: dict
    """

    parsed = urlparse(url)
    return dict(parse_qsl(parsed.query))


def get_domain(url: str) -> str:
    """
    Return SLD (Second Level Domain) from url
    :param url: str
    :return: str
    """

    return '.'.join(urlparse(url).netloc.split('.')[-2:])
